// File: DAGAd.h
// Author: Francesco Giacomini <Francesco.Giacomini@cnaf.infn.it>
// Copyright (c) 2001 EU DataGrid.
// For license conditions see http://www.eu-datagrid.org/license.html

// $Id: DAGAd.h,v 1.4 2003/03/13 18:18:28 giaco Exp $

#ifndef EDG_WORKLOAD_COMMON_REQUESTAD_DAGAD_H
#define EDG_WORKLOAD_COMMON_REQUESTAD_DAGAD_H

#include <string>
#include <utility>
#include <map>
#include <set>
#include <iosfwd>
#include <classad_distribution.h>

namespace edg {
namespace workload {
namespace common {
namespace requestad {

struct InvalidDAGInfo {};
struct InvalidNodeInfo {};

class DAGAdNodeIterator
  : public std::iterator<
  std::input_iterator_tag
  ,std::pair<std::string, classad::ExprTree const*>
  ,std::iterator_traits<classad::ClassAd::const_iterator>::difference_type
  ,std::pair<std::string, classad::ExprTree const*> const*
  ,std::pair<std::string, classad::ExprTree const*> const&
>
{
  friend bool operator==(DAGAdNodeIterator const& lhs, DAGAdNodeIterator const& rhs);

  classad::ClassAd const* m_nodes;
  classad::ClassAd::const_iterator m_node;
  value_type m_value;

  bool is_good() const
  {
    return m_node != m_nodes->end();
  }

  void set_value()
  {
    if (is_good()) {
      m_value = *m_node;
    }
  }

public:

  DAGAdNodeIterator()
    : m_nodes(0), m_node(), m_value()
  {
  }

  DAGAdNodeIterator(classad::ClassAd const* nodes, classad::ClassAd::const_iterator it)
    : m_nodes(nodes), m_node(it), m_value()
  {
    if (m_node != nodes->end()
        && m_node->second->GetKind() != classad::ExprTree::CLASSAD_NODE) {
      operator++();
    }
    set_value();
  }

  DAGAdNodeIterator& operator++()
  {
    if (m_node != m_nodes->end()) {
      do {
        ++m_node;
      } while (m_node != m_nodes->end()
               && m_node->second->GetKind() != classad::ExprTree::CLASSAD_NODE);
    }

    set_value();

    return *this;
  }

  DAGAdNodeIterator operator++(int)
  {
    DAGAdNodeIterator tmp(*this);
    ++*this;
    return tmp;
  }

  reference operator*() const
  {
    return m_value;
  }

  pointer operator->() const
  {
    return &(operator*());
  }

};

inline bool operator==(DAGAdNodeIterator const& lhs, DAGAdNodeIterator const& rhs)
{
  return lhs.m_nodes == rhs.m_nodes && lhs.m_node == rhs.m_node;
}

inline bool operator!=(DAGAdNodeIterator const& lhs, DAGAdNodeIterator const& rhs)
{
  return !(lhs == rhs);
}

struct DAGAdDependencyIterator
  : public std::iterator<
  std::input_iterator_tag
  ,std::pair<DAGAdNodeIterator, DAGAdNodeIterator>
  ,ptrdiff_t
  ,std::pair<DAGAdNodeIterator, DAGAdNodeIterator> const*
  ,std::pair<DAGAdNodeIterator, DAGAdNodeIterator> const&
>
{
  friend bool operator==(DAGAdDependencyIterator const& lhs,
                         DAGAdDependencyIterator const& rhs);

  typedef DAGAdNodeIterator node_iterator;
  typedef classad::ExprList::iterator Iterator;

  classad::ClassAd const* m_nodes;
  Iterator m_dep;
  Iterator m_dep_end;
  Iterator m_parent;
  Iterator m_child;
  value_type m_value;

  static classad::ExprList* parents(classad::ExprTree* dep)
  {
    return dynamic_cast<classad::ExprList*>(*dynamic_cast<classad::ExprList*>(dep)->begin());
  }
  static classad::ExprList* children(classad::ExprTree* dep)
  {
    return dynamic_cast<classad::ExprList*>(*++dynamic_cast<classad::ExprList*>(dep)->begin());
  }
  bool is_good() const { return m_dep != m_dep_end; }
  void set_value()
  {
    if (is_good()) {
      using namespace classad;

      ExprTree* t;
      std::string node1;
      std::string node2;
      bool a;
      dynamic_cast<AttributeReference const*>(*m_parent)->GetComponents(t, node1, a);
      dynamic_cast<AttributeReference const*>(*m_child)->GetComponents(t, node2, a);
      m_value = std::make_pair(
        node_iterator(m_nodes, m_nodes->find(node1)),
        node_iterator(m_nodes, m_nodes->find(node2))
      );
    }      
  }

  DAGAdDependencyIterator(classad::ClassAd const* nodes = 0, bool begin = false)
    : m_nodes(nodes), m_dep(), m_dep_end(), m_parent(), m_child()
  {
    if (nodes) {
      using namespace classad;
      ExprList* deps = dynamic_cast<ExprList*>(nodes->Lookup("dependencies"));

      if (begin) {
        m_dep = deps->begin();
        m_dep_end = deps->end();
        m_parent = parents(*m_dep)->begin();
        m_child = children(*m_dep)->begin();
        set_value();
      } else {
        m_dep = deps->end();
        m_dep_end = deps->end();
      }
    }
  }

  // copy ctor, operator=() and dtor are ok

  DAGAdDependencyIterator& operator++()
  {
    ++m_child;
    if (m_child == children(*m_dep)->end()) {
      ++m_parent;
      if (m_parent == parents(*m_dep)->end()) {
        ++m_dep;
        if (m_dep != m_dep_end) {
          m_parent = parents(*m_dep)->begin();
          m_child = children(*m_dep)->begin();
        }
      } else {
        m_child = children(*m_dep)->begin();
      }
    }

    set_value();

    return *this;
  }

  DAGAdDependencyIterator operator++(int)
  {
    DAGAdDependencyIterator tmp(*this);
    ++*this;
    return tmp;
  }

  reference operator*() const
  {
    return m_value;
  }

  pointer operator->() const
  {
    return &(operator*());
  }
};

inline bool operator==(DAGAdDependencyIterator const& lhs,
                       DAGAdDependencyIterator const& rhs)
{
  return (!lhs.is_good() && !rhs.is_good())
    || (lhs.is_good() && rhs.is_good()
        && lhs.m_dep == rhs.m_dep
        && lhs.m_parent == rhs.m_parent
        && lhs.m_child == rhs.m_child
        );
}

inline bool operator!=(DAGAdDependencyIterator const& lhs,
                DAGAdDependencyIterator const& rhs)
{
  return !(lhs == rhs);
}

class DAGAd
{
public:

  typedef DAGAdNodeIterator node_iterator;
  typedef DAGAdNodeIterator::value_type node_value_type;
  typedef DAGAdDependencyIterator dependency_iterator;
  typedef DAGAdDependencyIterator::value_type dependency_value_type;

private:
  classad::ClassAd m_ad;

public:
  DAGAd();
  explicit DAGAd(classad::ClassAd const& ad);

  classad::ClassAd const& ad() const;

  int max_nodes_running(int n);
  int max_nodes_running() const;
  std::string default_node_type(std::string const& type);
  std::string default_node_type() const;
  bool add_node(std::string const& name, classad::ClassAd const& info);
  bool remove_node(std::string const& name);
  bool add_dependency(std::string const& first, std::string const& second);
  bool remove_dependency(std::string const& first, std::string const& second);

  std::pair<dependency_iterator, dependency_iterator> dependencies() const;
  std::pair<node_iterator, node_iterator> nodes() const;
};

}}}} // edg::workload::common::requestad

#endif
